@@ -18,6 +18,13 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/*
+ * Fullscreen mode code adapted from gedit
+ *  Copyright (C) 1998, 1999 Alex Roberts, Evan Lawrence
+ *  Copyright (C) 2000, 2001 Chema Celorio, Paolo Maggi
+ *  Copyright (C) 2002-2005 Paolo Maggi
+ */
+
 #include "config.h"
 #include <string.h>
 #include <glib/gi18n-lib.h>
@@ -30,6 +37,8 @@
 #endif
 
 #include "dh-book-tree.h"
+#include "dh-book-manager.h"
+#include "dh-book.h"
 #include "dh-preferences.h"
 #include "dh-search.h"
 #include "dh-window.h"
@@ -39,6 +48,8 @@
 #include "eggfindbar.h"
 #include "ige-conf.h"
 
+#define FULLSCREEN_ANIMATION_SPEED 4
+
 struct _DhWindowPriv {
         DhBase         *base;
 
@@ -53,10 +64,15 @@ struct _DhWindowPriv {
         GtkWidget      *vbox;
         GtkWidget      *findbar;
 
+        GtkWidget      *fullscreen_controls;
+        guint           fullscreen_animation_timeout_id;
+        gboolean        fullscreen_animation_enter;
+
         GtkUIManager   *manager;
         GtkActionGroup *action_group;
 
         DhLink         *selected_search_link;
+        guint           find_source_id;
 };
 
 enum {
@@ -67,8 +83,8 @@ enum {
 static gint signals[LAST_SIGNAL] = { 0 };
 
 static guint tab_accel_keys[] = {
-        GDK_1, GDK_2, GDK_3, GDK_4, GDK_5,
-        GDK_6, GDK_7, GDK_8, GDK_9, GDK_0
+        GDK_KEY_1, GDK_KEY_2, GDK_KEY_3, GDK_KEY_4, GDK_KEY_5,
+        GDK_KEY_6, GDK_KEY_7, GDK_KEY_8, GDK_KEY_9, GDK_KEY_0
 };
 
 static const
@@ -94,6 +110,10 @@ zoom_levels[] =
 #define ZOOM_MAXIMAL    (zoom_levels[8].level)
 #define ZOOM_DEFAULT    (zoom_levels[2].level)
 
+#if GTK_CHECK_VERSION (2,17,5)
+#define ERRORS_IN_INFOBAR
+#endif
+
 static void           dh_window_class_init           (DhWindowClass   *klass);
 static void           dh_window_init                 (DhWindow        *window);
 static void           window_populate                (DhWindow        *window);
@@ -103,9 +123,6 @@ static void           window_tree_link_selected_cb   (GObject         *ignored,
 static void           window_search_link_selected_cb (GObject         *ignored,
                                                       DhLink          *link,
                                                       DhWindow        *window);
-static void           window_manager_add_widget      (GtkUIManager    *manager,
-                                                      GtkWidget       *widget,
-                                                      DhWindow        *window);
 static void           window_check_history           (DhWindow        *window,
                                                       WebKitWebView   *web_view);
 static void           window_web_view_tab_accel_cb   (GtkAccelGroup   *accel_group,
@@ -132,6 +149,9 @@ static int            window_open_new_tab            (DhWindow        *window,
                                                       const gchar     *location,
                                                       gboolean         switch_focus);
 static WebKitWebView *window_get_active_web_view     (DhWindow        *window);
+#ifdef ERRORS_IN_INFOBAR
+static GtkWidget *    window_get_active_info_bar     (DhWindow *window);
+#endif
 static void           window_update_title            (DhWindow        *window,
                                                       WebKitWebView   *web_view,
                                                       const gchar     *title);
@@ -362,6 +382,276 @@ window_activate_zoom_default (GtkAction *action,
         window_update_zoom_actions_sensitiveness (window);
 }
 
+static gboolean
+run_fullscreen_animation (gpointer data)
+{
+	DhWindow *window = DH_WINDOW (data);
+	GdkScreen *screen;
+	GdkRectangle fs_rect;
+	gint x, y;
+
+	screen = gtk_window_get_screen (GTK_WINDOW (window));
+	gdk_screen_get_monitor_geometry (screen,
+					 gdk_screen_get_monitor_at_window (screen,
+									   gtk_widget_get_window (GTK_WIDGET (window))),
+					 &fs_rect);
+
+	gtk_window_get_position (GTK_WINDOW (window->priv->fullscreen_controls),
+				 &x, &y);
+
+	if (window->priv->fullscreen_animation_enter)
+	{
+		if (y == fs_rect.y)
+		{
+			window->priv->fullscreen_animation_timeout_id = 0;
+			return FALSE;
+		}
+		else
+		{
+			gtk_window_move (GTK_WINDOW (window->priv->fullscreen_controls),
+					 x, y + 1);
+			return TRUE;
+		}
+	}
+	else
+	{
+		gint w, h;
+
+		gtk_window_get_size (GTK_WINDOW (window->priv->fullscreen_controls),
+				     &w, &h);
+
+		if (y == fs_rect.y - h + 1)
+		{
+			window->priv->fullscreen_animation_timeout_id = 0;
+			return FALSE;
+		}
+		else
+		{
+			gtk_window_move (GTK_WINDOW (window->priv->fullscreen_controls),
+					 x, y - 1);
+			return TRUE;
+		}
+	}
+}
+
+static void
+show_hide_fullscreen_toolbar (DhWindow *window,
+                              gboolean     show,
+                              gint         height)
+{
+        GtkSettings *settings;
+        gboolean enable_animations;
+
+        settings = gtk_widget_get_settings (GTK_WIDGET (window));
+        g_object_get (G_OBJECT (settings),
+                      "gtk-enable-animations",
+                      &enable_animations,
+                      NULL);
+
+        if (enable_animations)
+        {
+                window->priv->fullscreen_animation_enter = show;
+
+                if (window->priv->fullscreen_animation_timeout_id == 0)
+                {
+                        window->priv->fullscreen_animation_timeout_id =
+                                g_timeout_add (FULLSCREEN_ANIMATION_SPEED,
+                                               (GSourceFunc) run_fullscreen_animation,
+                                               window);
+                }
+        }
+        else
+        {
+                GdkRectangle fs_rect;
+                GdkScreen *screen;
+
+                screen = gtk_window_get_screen (GTK_WINDOW (window));
+                gdk_screen_get_monitor_geometry (screen,
+                                                 gdk_screen_get_monitor_at_window (screen,
+                                                                                   gtk_widget_get_window (GTK_WIDGET (window))),
+                                                 &fs_rect);
+
+                if (show)
+                        gtk_window_move (GTK_WINDOW (window->priv->fullscreen_controls),
+                                 fs_rect.x, fs_rect.y);
+                else
+                        gtk_window_move (GTK_WINDOW (window->priv->fullscreen_controls),
+                                         fs_rect.x, fs_rect.y - height + 1);
+        }
+
+}
+
+
+static gboolean
+on_fullscreen_controls_enter_notify_event (GtkWidget        *widget,
+                                           GdkEventCrossing *event,
+                                           DhWindow      *window)
+{
+        show_hide_fullscreen_toolbar (window, TRUE, 0);
+
+        return FALSE;
+}
+
+static gboolean
+on_fullscreen_controls_leave_notify_event (GtkWidget        *widget,
+                                           GdkEventCrossing *event,
+                                           DhWindow      *window)
+{
+        GdkDisplay *display;
+        GdkScreen *screen;
+        gint w, h;
+        gint x, y;
+
+        display = gdk_display_get_default ();
+        screen = gtk_window_get_screen (GTK_WINDOW (window));
+
+        gtk_window_get_size (GTK_WINDOW (window->priv->fullscreen_controls), &w, &h);
+        gdk_display_get_pointer (display, &screen, &x, &y, NULL);
+
+        /* gtk seems to emit leave notify when clicking on tool items,
+         * work around it by checking the coordinates
+         */
+        if (y >= h)
+        {
+                show_hide_fullscreen_toolbar (window, FALSE, h);
+        }
+
+        return FALSE;
+}
+
+static gboolean
+window_is_fullscreen (DhWindow *window)
+{
+        GdkWindowState  state;
+
+        g_return_val_if_fail (DH_IS_WINDOW (window), FALSE);
+
+#if GTK_CHECK_VERSION (2,14,0)
+        state = gdk_window_get_state (gtk_widget_get_window (GTK_WIDGET (window)));
+#else
+        state = gdk_window_get_state (GTK_WIDGET (window)->window);
+#endif
+
+        return state & GDK_WINDOW_STATE_FULLSCREEN;
+}
+
+static void
+window_fullscreen_controls_build (DhWindow *window)
+{
+        GtkWidget *toolbar;
+        GtkAction *action;
+        DhWindowPriv  *priv;
+
+        priv = window->priv;
+        if (priv->fullscreen_controls != NULL)
+                return;
+
+        priv->fullscreen_controls = gtk_window_new (GTK_WINDOW_POPUP);
+        gtk_window_set_transient_for (GTK_WINDOW (priv->fullscreen_controls),
+                                      GTK_WINDOW (window));
+
+        toolbar = gtk_ui_manager_get_widget (priv->manager, "/FullscreenToolBar");
+        gtk_container_add (GTK_CONTAINER (priv->fullscreen_controls),
+                           toolbar);
+        action = gtk_action_group_get_action (priv->action_group,
+                                              "LeaveFullscreen");
+        g_object_set (action, "is-important", TRUE, NULL);
+
+        /* Set the toolbar style */
+        gtk_toolbar_set_style (GTK_TOOLBAR (toolbar),
+                               GTK_TOOLBAR_BOTH_HORIZ);
+
+        g_signal_connect (priv->fullscreen_controls, "enter-notify-event",
+                          G_CALLBACK (on_fullscreen_controls_enter_notify_event),
+                          window);
+        g_signal_connect (priv->fullscreen_controls, "leave-notify-event",
+                          G_CALLBACK (on_fullscreen_controls_leave_notify_event),
+                          window);
+}
+
+static void
+window_fullscreen_controls_show (DhWindow *window)
+{
+        GdkScreen *screen;
+        GdkRectangle fs_rect;
+        gint w, h;
+
+        screen = gtk_window_get_screen (GTK_WINDOW (window));
+        gdk_screen_get_monitor_geometry (screen,
+                        gdk_screen_get_monitor_at_window (
+                                screen,
+                                gtk_widget_get_window (GTK_WIDGET (window))),
+                        &fs_rect);
+
+        gtk_window_get_size (GTK_WINDOW (window->priv->fullscreen_controls), &w, &h);
+
+        gtk_window_resize (GTK_WINDOW (window->priv->fullscreen_controls),
+                           fs_rect.width, h);
+
+        gtk_window_move (GTK_WINDOW (window->priv->fullscreen_controls),
+                         fs_rect.x, fs_rect.y - h + 1);
+
+        gtk_widget_show_all (window->priv->fullscreen_controls);
+}
+
+static void
+window_fullscreen (DhWindow *window)
+{
+        if (window_is_fullscreen (window))
+                return;
+
+        gtk_window_fullscreen (GTK_WINDOW (window));
+        gtk_widget_hide (gtk_ui_manager_get_widget (window->priv->manager, "/MenuBar"));
+        gtk_widget_hide (gtk_ui_manager_get_widget (window->priv->manager, "/Toolbar"));
+
+        window_fullscreen_controls_build (window);
+        window_fullscreen_controls_show (window);
+}
+
+static void
+window_unfullscreen (DhWindow *window)
+{
+        if (! window_is_fullscreen (window))
+                return;
+
+        gtk_window_unfullscreen (GTK_WINDOW (window));
+        gtk_widget_show (gtk_ui_manager_get_widget (window->priv->manager, "/MenuBar"));
+        gtk_widget_show (gtk_ui_manager_get_widget (window->priv->manager, "/Toolbar"));
+
+        gtk_widget_hide (window->priv->fullscreen_controls);
+}
+
+
+static void
+window_toggle_fullscreen_mode (GtkAction *action,
+                               DhWindow  *window)
+{
+        if (window_is_fullscreen (window)) {
+                window_unfullscreen (window);
+        } else {
+                window_fullscreen (window);
+        }
+}
+
+static void
+window_leave_fullscreen_mode (GtkAction *action,
+                              DhWindow *window)
+{
+        GtkAction *view_action;
+
+        view_action = gtk_action_group_get_action (window->priv->action_group,
+                                                   "ViewFullscreen");
+        g_signal_handlers_block_by_func (view_action,
+                        G_CALLBACK (window_toggle_fullscreen_mode),
+                        window);
+        gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (view_action),
+                                      FALSE);
+        window_unfullscreen (window);
+        g_signal_handlers_unblock_by_func (view_action,
+                        G_CALLBACK (window_toggle_fullscreen_mode),
+                        window);
+}
+
 static void
 window_activate_preferences (GtkAction *action,
                              DhWindow  *window)
@@ -438,6 +728,7 @@ window_activate_about (GtkAction *action,
                 "Richard Hult <richard@imendio.com>",
                 "Johan Dahlin <johan@gnome.org>",
                 "Ross Burton <ross@burtonini.com>",
+                "Aleksander Morgado <aleksander@lanedo.com>",
                 NULL
         };
         const gchar **documenters = NULL;
@@ -448,14 +739,15 @@ window_activate_about (GtkAction *action,
         gtk_show_about_dialog (GTK_WINDOW (window),
                                "name", _("Devhelp"),
                                "version", PACKAGE_VERSION,
-                               "comments", _("A developer's help browser for GNOME"),
+                               "comments", _("A developers' help browser for GNOME"),
                                "authors", authors,
                                "documenters", documenters,
                                "translator-credits",
                                strcmp (translator_credits, "translator_credits") != 0 ?
                                translator_credits : NULL,
-                               "website", "http://live.gnome.org/devhelp",
-                               "logo-icon-name", "devhelp",
+                               "website", PACKAGE_URL,
+                               "website-label", _("DevHelp Website"),
+                               "logo-icon-name", PACKAGE_TARNAME,
                                NULL);
 }
 
@@ -489,7 +781,7 @@ static const GtkActionEntry actions[] = {
           G_CALLBACK (window_activate_new_window) },
         { "NewTab", GTK_STOCK_NEW, N_("New _Tab"), "<control>T", NULL,
           G_CALLBACK (window_activate_new_tab) },
-        { "Print", GTK_STOCK_PRINT, N_("_Print..."), "<control>P", NULL,
+        { "Print", GTK_STOCK_PRINT, N_("_Print…"), "<control>P", NULL,
           G_CALLBACK (window_activate_print) },
         { "Close", GTK_STOCK_CLOSE, NULL, NULL, NULL,
           G_CALLBACK (window_activate_close) },
@@ -536,6 +828,18 @@ static const GtkActionEntry actions[] = {
         /* About menu */
         { "About", GTK_STOCK_ABOUT, NULL, NULL, NULL,
           G_CALLBACK (window_activate_about) },
+
+        /* Fullscreen toolbar */
+        { "LeaveFullscreen", GTK_STOCK_LEAVE_FULLSCREEN, NULL,
+          NULL, N_("Leave fullscreen mode"),
+          G_CALLBACK (window_leave_fullscreen_mode) }
+};
+
+static const GtkToggleActionEntry always_sensitive_toggle_menu_entries[] =
+{
+        { "ViewFullscreen", GTK_STOCK_FULLSCREEN, NULL, "F11",
+          N_("Display in full screen"),
+          G_CALLBACK (window_toggle_fullscreen_mode), FALSE },
 };
 
 static const gchar* important_actions[] = {
@@ -604,10 +908,10 @@ dh_window_init (DhWindow *window)
         accel_group = gtk_ui_manager_get_accel_group (priv->manager);
         gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
 
-        priv->main_box = gtk_vbox_new (FALSE, 0);
+        priv->main_box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
         gtk_widget_show (priv->main_box);
 
-        priv->menu_box = gtk_vbox_new (FALSE, 0);
+        priv->menu_box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
         gtk_widget_show (priv->menu_box);
         gtk_container_set_border_width (GTK_CONTAINER (priv->menu_box), 0);
         gtk_box_pack_start (GTK_BOX (priv->main_box), priv->menu_box,
@@ -620,11 +924,6 @@ dh_window_init (DhWindow *window)
                           G_CALLBACK (window_open_link_cb),
                           window);
 
-        g_signal_connect (priv->manager,
-                          "add-widget",
-                          G_CALLBACK (window_manager_add_widget),
-                          window);
-
         priv->action_group = gtk_action_group_new ("MainWindow");
 
         gtk_action_group_set_translation_domain (priv->action_group,
@@ -634,6 +933,10 @@ dh_window_init (DhWindow *window)
                                       actions,
                                       G_N_ELEMENTS (actions),
                                       window);
+        gtk_action_group_add_toggle_actions (priv->action_group,
+                                             always_sensitive_toggle_menu_entries,
+                                             G_N_ELEMENTS (always_sensitive_toggle_menu_entries),
+                                             window);
 
         for (i = 0; i < G_N_ELEMENTS (important_actions); i++) {
                 action = gtk_action_group_get_action (priv->action_group,
@@ -681,7 +984,7 @@ dh_window_init (DhWindow *window)
  */
 static void
 window_control_switch_page_cb (GtkWidget       *notebook,
-                               GtkNotebookPage *page,
+                               gpointer         page,
                                guint            page_num,
                                DhWindow        *window)
 {
@@ -696,7 +999,7 @@ window_control_switch_page_cb (GtkWidget       *notebook,
 
 static void
 window_control_after_switch_page_cb (GtkWidget       *notebook,
-                                     GtkNotebookPage *page,
+                                     gpointer         page,
                                      guint            page_num,
                                      DhWindow        *window)
 {
@@ -711,7 +1014,7 @@ window_control_after_switch_page_cb (GtkWidget       *notebook,
 
 static void
 window_web_view_switch_page_cb (GtkNotebook     *notebook,
-                                GtkNotebookPage *page,
+                                gpointer         page,
                                 guint            new_page_num,
                                 DhWindow        *window)
 {
@@ -749,7 +1052,7 @@ window_web_view_switch_page_cb (GtkNotebook     *notebook,
 
 static void
 window_web_view_switch_page_after_cb (GtkNotebook     *notebook,
-                                      GtkNotebookPage *page,
+                                      gpointer         page,
                                       guint            new_page_num,
                                       DhWindow        *window)
 {
@@ -759,11 +1062,12 @@ window_web_view_switch_page_after_cb (GtkNotebook     *notebook,
 static void
 window_populate (DhWindow *window)
 {
-        DhWindowPriv *priv;
-        gchar        *path;
-        GtkWidget    *book_tree_sw;
-        GNode        *contents_tree;
-        GList        *keywords;
+        DhWindowPriv  *priv;
+        gchar         *path;
+        GtkWidget     *book_tree_sw;
+        DhBookManager *book_manager;
+        GtkWidget     *menubar;
+        GtkWidget     *toolbar;
 
         priv = window->priv;
 
@@ -774,19 +1078,27 @@ window_populate (DhWindow *window)
         g_free (path);
         gtk_ui_manager_ensure_update (priv->manager);
 
+        menubar = gtk_ui_manager_get_widget (priv->manager, "/MenuBar");
+        gtk_box_pack_start (GTK_BOX (priv->menu_box), menubar,
+                            FALSE, FALSE, 0);
+        toolbar = gtk_ui_manager_get_widget (priv->manager, "/Toolbar");
+        gtk_box_pack_start (GTK_BOX (priv->menu_box), toolbar,
+                            FALSE, FALSE, 0);
+
+	gtk_style_context_add_class (gtk_widget_get_style_context (toolbar),
+				     GTK_STYLE_CLASS_PRIMARY_TOOLBAR);
+
 #ifdef GDK_WINDOWING_QUARTZ
         {
                 GtkWidget       *widget;
                 IgeMacMenuGroup *group;
 
                 /* Hide toolbar labels. */
-                widget = gtk_ui_manager_get_widget (priv->manager, "/Toolbar");
-                gtk_toolbar_set_style (GTK_TOOLBAR (widget), GTK_TOOLBAR_ICONS);
+                gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_ICONS);
 
                 /* Setup menubar. */
-                widget = gtk_ui_manager_get_widget (priv->manager, "/MenuBar");
-                ige_mac_menu_set_menu_bar (GTK_MENU_SHELL (widget));
-                gtk_widget_hide (widget);
+                ige_mac_menu_set_menu_bar (GTK_MENU_SHELL (menubar));
+                gtk_widget_hide (menubar);
 
                 widget = gtk_ui_manager_get_widget (priv->manager, "/MenuBar/FileMenu/Quit");
                 ige_mac_menu_set_quit_menu_item (GTK_MENU_ITEM (widget));
@@ -802,13 +1114,13 @@ window_populate (DhWindow *window)
                 group =  ige_mac_menu_add_app_menu_group ();
                 widget = gtk_ui_manager_get_widget (priv->manager, "/MenuBar/EditMenu/Preferences");
                 ige_mac_menu_add_app_menu_item (group, GTK_MENU_ITEM (widget),
-                                                _("Preferences..."));
+                                                _("Preferences…"));
 
                 ige_mac_menu_set_global_key_handler_enabled (TRUE);
         }
 #endif
 
-        priv->hpaned = gtk_hpaned_new ();
+        priv->hpaned = gtk_paned_new (GTK_ORIENTATION_HORIZONTAL);
 
         gtk_box_pack_start (GTK_BOX (priv->main_box), priv->hpaned, TRUE, TRUE, 0);
 
@@ -836,10 +1148,9 @@ window_populate (DhWindow *window)
                                              GTK_SHADOW_IN);
         gtk_container_set_border_width (GTK_CONTAINER (book_tree_sw), 2);
 
-        contents_tree = dh_base_get_book_tree (priv->base);
-        keywords = dh_base_get_keywords (priv->base);
+        book_manager = dh_base_get_book_manager (priv->base);
 
-        priv->book_tree = dh_book_tree_new (contents_tree);
+        priv->book_tree = dh_book_tree_new (book_manager);
         gtk_container_add (GTK_CONTAINER (book_tree_sw),
                            priv->book_tree);
         dh_util_state_set_notebook_page_name (book_tree_sw, "content");
@@ -851,7 +1162,7 @@ window_populate (DhWindow *window)
                           G_CALLBACK (window_tree_link_selected_cb),
                           window);
 
-        priv->search = dh_search_new (keywords);
+        priv->search = dh_search_new (book_manager);
         dh_util_state_set_notebook_page_name (priv->search, "search");
         gtk_notebook_append_page (GTK_NOTEBOOK (priv->control_notebook),
                                   priv->search,
@@ -861,7 +1172,7 @@ window_populate (DhWindow *window)
                           G_CALLBACK (window_search_link_selected_cb),
                           window);
 
-        priv->vbox = gtk_vbox_new (FALSE, 0);
+        priv->vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
         gtk_paned_add2 (GTK_PANED (priv->hpaned), priv->vbox);
 
         /* HTML tabs notebook. */
@@ -914,7 +1225,7 @@ window_populate (DhWindow *window)
 }
 
 
-static gchar*
+static gchar *
 find_library_equivalent (DhWindow    *window,
                          const gchar *uri)
 {
@@ -922,26 +1233,38 @@ find_library_equivalent (DhWindow    *window,
         gchar **components;
         GList *iter;
         DhLink *link;
+        DhBookManager *book_manager;
         gchar *book_id;
         gchar *filename;
         gchar *local_uri = NULL;
+        GList *books;
 
         components = g_strsplit (uri, "/", 0);
         book_id = components[4];
         filename = components[6];
 
         priv = window->priv;
+        book_manager = dh_base_get_book_manager (priv->base);
 
-        for (iter = dh_base_get_keywords (priv->base); iter; iter = g_list_next (iter)) {
-                link = iter->data;
-                if (g_strcmp0 (dh_link_get_book_id (link), book_id) != 0) {
-                        continue;
+        /* use list pointer to iterate */
+        for (books = dh_book_manager_get_books (book_manager);
+             !local_uri && books;
+             books = g_list_next (books)) {
+                DhBook *book = DH_BOOK (books->data);
+
+                for (iter = dh_book_get_keywords (book);
+                     iter;
+                     iter = g_list_next (iter)) {
+                        link = iter->data;
+                        if (g_strcmp0 (dh_link_get_book_id (link), book_id) != 0) {
+                                continue;
+                        }
+                        if (g_strcmp0 (dh_link_get_file_name (link), filename) != 0) {
+                                continue;
+                        }
+                        local_uri = dh_link_get_uri (link);
+                        break;
                 }
-                if (g_strcmp0 (dh_link_get_file_name (link), filename) != 0) {
-                        continue;
-                }
-                local_uri = dh_link_get_uri (link);
-                break;
         }
 
         g_strfreev (components);
@@ -965,6 +1288,11 @@ window_web_view_navigation_policy_decision_requested (WebKitWebView
 
         uri = webkit_network_request_get_uri (request);
 
+#ifdef ERRORS_IN_INFOBAR
+        /* make sure to hide the info bar on page change */
+        gtk_widget_hide (window_get_active_info_bar (window));
+#endif
+
         if (webkit_web_navigation_action_get_button (navigation_action) == 2) { /* middle click */
                 webkit_web_policy_decision_ignore (policy_decision);
                 g_signal_emit (window, signals[OPEN_LINK], 0, uri, DH_OPEN_LINK_NEW_TAB);
@@ -999,6 +1327,45 @@ window_web_view_navigation_policy_decision_requested (WebKitWebView
         return FALSE;
 }
 
+
+static gboolean
+window_web_view_load_error_cb (WebKitWebView  *web_view,
+                               WebKitWebFrame *frame,
+                               gchar          *uri,
+                               gpointer       *web_error,
+                               DhWindow       *window)
+{
+#ifdef ERRORS_IN_INFOBAR
+        GtkWidget *info_bar;
+        GtkWidget *content_area;
+        GtkWidget *message_label;
+        GList     *children;
+        gchar     *markup;
+
+        info_bar = window_get_active_info_bar (window);
+        markup = g_strdup_printf ("<b>%s</b>",
+                       _("Error opening the requested link."));
+        message_label = gtk_label_new (markup);
+        gtk_misc_set_alignment (GTK_MISC (message_label), 0, 0.5);
+        gtk_label_set_use_markup (GTK_LABEL (message_label), TRUE);
+        content_area = gtk_info_bar_get_content_area (GTK_INFO_BAR (info_bar));
+        children = gtk_container_get_children (GTK_CONTAINER (content_area));
+        if (children) {
+                gtk_container_remove (GTK_CONTAINER (content_area), children->data);
+                g_list_free (children);
+        }
+        gtk_container_add (GTK_CONTAINER (content_area), message_label);
+        gtk_widget_show (message_label);
+
+        gtk_widget_show (info_bar);
+        g_free (markup);
+
+        return TRUE;
+#else
+	return FALSE;
+#endif
+}
+
 static void
 window_tree_link_selected_cb (GObject  *ignored,
                               DhLink   *link,
@@ -1013,7 +1380,7 @@ window_tree_link_selected_cb (GObject  *ignored,
         view = window_get_active_web_view (window);
 
         uri = dh_link_get_uri (link);
-        webkit_web_view_open (view, uri);
+        webkit_web_view_load_uri (view, uri);
         g_free (uri);
 
         window_check_history (window, view);
@@ -1035,27 +1402,12 @@ window_search_link_selected_cb (GObject  *ignored,
         view = window_get_active_web_view (window);
 
         uri = dh_link_get_uri (link);
-        webkit_web_view_open (view, uri);
+        webkit_web_view_load_uri (view, uri);
         g_free (uri);
 
         window_check_history (window, view);
 }
 
-static void
-window_manager_add_widget (GtkUIManager *manager,
-                           GtkWidget    *widget,
-                           DhWindow     *window)
-{
-        DhWindowPriv *priv;
-
-        priv = window->priv;
-
-        gtk_box_pack_start (GTK_BOX (priv->menu_box), widget,
-                            FALSE, FALSE, 0);
-
-        gtk_widget_show (widget);
-}
-
 static void
 window_check_history (DhWindow      *window,
                       WebKitWebView *web_view)
@@ -1101,15 +1453,14 @@ window_web_view_button_press_event_cb (WebKitWebView  *web_view,
         return FALSE;
 }
 
-static void
-window_find_search_changed_cb (GObject    *object,
-                               GParamSpec *pspec,
-                               DhWindow   *window)
+static gboolean
+do_search (DhWindow *window)
 {
-        DhWindowPriv  *priv;
+        DhWindowPriv  *priv = window->priv;
         WebKitWebView *web_view;
 
-        priv = window->priv;
+        priv->find_source_id = 0;
+
         web_view = window_get_active_web_view (window);
 
         webkit_web_view_unmark_text_matches (web_view);
@@ -1123,6 +1474,23 @@ window_find_search_changed_cb (GObject    *object,
                 web_view, egg_find_bar_get_search_string (EGG_FIND_BAR (priv->findbar)),
                 egg_find_bar_get_case_sensitive (EGG_FIND_BAR (priv->findbar)),
                 TRUE, TRUE);
+
+	return FALSE;
+}
+
+static void
+window_find_search_changed_cb (GObject    *object,
+                               GParamSpec *pspec,
+                               DhWindow   *window)
+{
+        DhWindowPriv *priv = window->priv;
+
+        if (priv->find_source_id != 0) {
+                g_source_remove (priv->find_source_id);
+                priv->find_source_id = 0;
+        }
+
+        priv->find_source_id = g_timeout_add (300, (GSourceFunc)do_search, window);
 }
 
 static void
@@ -1240,17 +1608,33 @@ window_open_new_tab (DhWindow    *window,
 {
         DhWindowPriv *priv;
         GtkWidget    *view;
+        GtkWidget    *vbox;
         GtkWidget    *scrolled_window;
         GtkWidget    *label;
         gint          num;
+#ifdef ERRORS_IN_INFOBAR
+        GtkWidget    *info_bar;
+#endif
 
         priv = window->priv;
 
+        /* Prepare the web view */
         view = webkit_web_view_new ();
         gtk_widget_show (view);
-
         dh_util_font_add_web_view (WEBKIT_WEB_VIEW (view));
 
+#ifdef ERRORS_IN_INFOBAR
+        /* Prepare the info bar */
+        info_bar = gtk_info_bar_new ();
+        gtk_widget_set_no_show_all (info_bar, TRUE);
+        gtk_info_bar_add_button (GTK_INFO_BAR (info_bar),
+                                 GTK_STOCK_CLOSE, GTK_RESPONSE_OK);
+        gtk_info_bar_set_message_type (GTK_INFO_BAR (info_bar),
+                                       GTK_MESSAGE_ERROR);
+        g_signal_connect (info_bar, "response",
+                          G_CALLBACK (gtk_widget_hide), NULL);
+#endif
+
 #if 0
         /* Leave this in for now to make it easier to experiment. */
         {
@@ -1263,15 +1647,26 @@ window_open_new_tab (DhWindow    *window,
         }
 #endif
 
+        vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
+        gtk_widget_show (vbox);
+
+        /* XXX: Really it would be much better to use real structures */
+        g_object_set_data (G_OBJECT (vbox), "web_view", view);
+#ifdef ERRORS_IN_INFOBAR
+        g_object_set_data (G_OBJECT (vbox), "info_bar", info_bar);
+
+        gtk_box_pack_start (GTK_BOX(vbox), info_bar, FALSE, TRUE, 0);
+#endif
+
         scrolled_window = gtk_scrolled_window_new (NULL, NULL);
         gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                         GTK_POLICY_AUTOMATIC,
                                         GTK_POLICY_AUTOMATIC);
-        gtk_widget_show (scrolled_window);
         gtk_container_add (GTK_CONTAINER (scrolled_window), view);
-        g_object_set_data (G_OBJECT (scrolled_window), "web_view", view);
+        gtk_widget_show (scrolled_window);
+        gtk_box_pack_start (GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);
 
-        label = window_new_tab_label (window, _("Empty Page"), scrolled_window);
+        label = window_new_tab_label (window, _("Empty Page"), vbox);
         gtk_widget_show_all (label);
 
         g_signal_connect (view, "title-changed",
@@ -1280,16 +1675,18 @@ window_open_new_tab (DhWindow    *window,
         g_signal_connect (view, "button-press-event",
                           G_CALLBACK (window_web_view_button_press_event_cb),
                           window);
-
         g_signal_connect (view, "navigation-policy-decision-requested",
                           G_CALLBACK (window_web_view_navigation_policy_decision_requested),
                           window);
+        g_signal_connect (view, "load-error",
+                          G_CALLBACK (window_web_view_load_error_cb),
+                          window);
 
         num = gtk_notebook_append_page (GTK_NOTEBOOK (priv->notebook),
-                                        scrolled_window, NULL);
+                                        vbox, NULL);
 
         gtk_notebook_set_tab_label (GTK_NOTEBOOK (priv->notebook),
-                                    scrolled_window, label);
+                                    vbox, label);
 
         if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (priv->notebook)) > 1) {
                 gtk_notebook_set_show_tabs (GTK_NOTEBOOK (priv->notebook), TRUE);
@@ -1300,7 +1697,7 @@ window_open_new_tab (DhWindow    *window,
         if (location) {
                 webkit_web_view_load_uri (WEBKIT_WEB_VIEW (view), location);
         } else {
-                webkit_web_view_open (WEBKIT_WEB_VIEW (view), "about:blank");
+                webkit_web_view_load_uri (WEBKIT_WEB_VIEW (view), "about:blank");
         }
 
         if (switch_focus) {
@@ -1374,7 +1771,7 @@ window_new_tab_label (DhWindow        *window,
         GtkWidget *close_button;
         GtkWidget *image;
 
-        hbox = gtk_hbox_new (FALSE, 4);
+        hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 4);
 
         label = gtk_label_new (str);
         gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
@@ -1431,6 +1828,27 @@ window_get_active_web_view (DhWindow *window)
         return g_object_get_data (G_OBJECT (page), "web_view");
 }
 
+#ifdef ERRORS_IN_INFOBAR
+static GtkWidget *
+window_get_active_info_bar (DhWindow *window)
+{
+        DhWindowPriv *priv;
+        gint          page_num;
+        GtkWidget    *page;
+
+        priv = window->priv;
+
+        page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook));
+        if (page_num == -1) {
+                return NULL;
+        }
+
+        page = gtk_notebook_get_nth_page (GTK_NOTEBOOK (priv->notebook), page_num);
+
+        return g_object_get_data (G_OBJECT (page), "info_bar");
+}
+#endif
+
 static void
 window_update_title (DhWindow      *window,
                      WebKitWebView *web_view,
@@ -1485,6 +1903,7 @@ window_tab_set_title (DhWindow      *window,
         GtkWidget    *page;
         GtkWidget    *hbox;
         GtkWidget    *label;
+        GtkWidget    *page_web_view;
 
         priv = window->priv;
 
@@ -1496,9 +1915,10 @@ window_tab_set_title (DhWindow      *window,
         for (i = 0; i < num_pages; i++) {
                 page = gtk_notebook_get_nth_page (
                         GTK_NOTEBOOK (priv->notebook), i);
+                page_web_view = g_object_get_data (G_OBJECT (page), "web_view");
 
                 /* The web_view widget is inside a frame. */
-                if (gtk_bin_get_child (GTK_BIN (page)) == GTK_WIDGET (web_view)) {
+                if (page_web_view == GTK_WIDGET (web_view)) {
                         hbox = gtk_notebook_get_tab_label (
                                 GTK_NOTEBOOK (priv->notebook), page);
 
@@ -1537,8 +1957,7 @@ dh_window_new (DhBase *base)
 
 void
 dh_window_search (DhWindow    *window,
-                  const gchar *str,
-                  const gchar *book_id)
+                  const gchar *str)
 {
         DhWindowPriv *priv;
 
@@ -1547,7 +1966,7 @@ dh_window_search (DhWindow    *window,
         priv = window->priv;
 
         gtk_notebook_set_current_page (GTK_NOTEBOOK (priv->control_notebook), 1);
-        dh_search_set_search_string (DH_SEARCH (priv->search), str, book_id);
+        dh_search_set_search_string (DH_SEARCH (priv->search), str, NULL);
 }
 
 void
@@ -1577,6 +1996,6 @@ _dh_window_display_uri (DhWindow    *window,
         priv = window->priv;
 
         web_view = window_get_active_web_view (window);
-        webkit_web_view_open (web_view, uri);
+        webkit_web_view_load_uri (web_view, uri);
         dh_book_tree_select_uri (DH_BOOK_TREE (priv->book_tree), uri);
 }
